// filesystem_storage_plugin.cpp
//
// Copyright (C) 2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#include <iostream>
#include <cctype>

#if defined(_WIN32) && !defined(USE_LIBXML2)
#	import <msxml4.dll>
#	pragma warning( push )
#	pragma warning( disable : 4290 )
#	include <atlbase.h>
#	pragma warning( pop )
#else
#	include <libxml/tree.h>
#	include <libxml/parser.h>
#endif

#include <openassetlib/plugins/filesystem_storage_plugin/filesystem_storage_plugin.hpp>
#include <openassetlib/plugins/filesystem_storage_plugin/query.hpp>

namespace olib { namespace openassetlib { namespace plugins { namespace filesystem {

filesystem_storage_plugin::filesystem_storage_plugin(void)
{
#if defined(_WIN32) && !defined(USE_LIBXML2)
	::CoInitialize(0);
#endif
}

filesystem_storage_plugin::~filesystem_storage_plugin(void)
{
#if defined(_WIN32) && !defined(USE_LIBXML2)
	::CoUninitialize();
#endif
}

al::set_ptr filesystem_storage_plugin::store(const al::set_ptr& /*set*/)
{
	return al::set_ptr();
}

al::set_ptr filesystem_storage_plugin::load(const opl::string& query_xml)
{
	const char* XML_ROOT_ELEMENT_NAME		= "filesystem_query";
	const char* XML_ELEMENT_FILESIZE		= "filesize";
	const char* XML_ELEMENT_LASTMODIFIED	= "lastmodified";
	const char* XML_ATTRIBUTE_SEARCHFOLDER	= "searchfolder";
	const char* XML_ATTRIBUTE_RECURSE		= "recurse";
	const char* XML_ATTRIBUTE_FILENAMEREGEX	= "filenameregex";
	const char* XML_ATTRIBUTE_COMPARITOR	= "comparitor";
	const char* XML_ATTRIBUTE_YEAR			= "year";
	const char* XML_ATTRIBUTE_MONTH			= "month";
	const char* XML_ATTRIBUTE_DAY			= "day";
	const char* XML_ATTRIBUTE_HOUR			= "hour";
	const char* XML_ATTRIBUTE_MINUTE		= "minute";
	const char* XML_ATTRIBUTE_SECOND		= "second";

	const char* TRUE_STRING					= "TRUE";
	const char* FALSE_STRING				= "FALSE";

#if defined(_WIN32) && !defined(USE_LIBXML2) //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	try
	{
		// Create XML DOM and load up with string passed in
		MSXML2::IXMLDOMDocument2Ptr pXMLDoc(__uuidof(MSXML2::DOMDocument40));
		pXMLDoc->validateOnParse = VARIANT_FALSE;
		if(pXMLDoc->loadXML(_bstr_t(query_xml.c_str())) == VARIANT_FALSE)
			return al::set_ptr();

		// Get the root element
		MSXML2::IXMLDOMElementPtr pElement = pXMLDoc->documentElement;
		if(pElement == NULL)
			return al::set_ptr();

		// Check root elements name
		if(pElement->nodeName != _bstr_t(XML_ROOT_ELEMENT_NAME))
			return al::set_ptr();

		// Get attribute node map
		MSXML2::IXMLDOMNamedNodeMapPtr pNodeMap = pElement->attributes;
		if(pNodeMap == NULL)
			return al::set_ptr();

		// Read search folder attribute
		MSXML2::IXMLDOMNodePtr pNode = pNodeMap->getNamedItem(_bstr_t(XML_ATTRIBUTE_SEARCHFOLDER));
		if(pNode == NULL)
			return al::set_ptr();

		fs::path root_folder(pNode->text, fs::native);

		// Read recurse attribute
		bool recurse = false;
		pNode = pNodeMap->getNamedItem(_bstr_t(XML_ATTRIBUTE_RECURSE));
		if(pNode != NULL)
		{
			std::string text(pNode->text);
			std::transform(text.begin(), text.end(), text.begin(), (int(*)(int))std::toupper);
			if(text.compare(TRUE_STRING) == 0)
				recurse = true;
			else if(text.compare(FALSE_STRING) == 0)
				recurse = false;
			else
				return al::set_ptr();
		}

		// Read filename regular expression
		pNode = pNodeMap->getNamedItem(_bstr_t(XML_ATTRIBUTE_FILENAMEREGEX));
		if(pNode == NULL)
			return al::set_ptr();

		opl::string filename_regex(pNode->text);

		// Process optional elements...
		opl::string filesize_comparitor;
		opl::string datetime_comparitor;
		std::tm comparison_datetime;
		long long comparison_filesize	= 0;
		bool filesize_option			= false;
		bool lastmodified_option		= false;

		// Process optional filesize child element
		MSXML2::IXMLDOMNodeListPtr pNodeList = pElement->getElementsByTagName(_bstr_t(XML_ELEMENT_FILESIZE));
		if(pNodeList == NULL)
			return al::set_ptr();

		if(pNodeList->length > 0)
		{
			filesize_option = true;

			pNode = pNodeList->item[0];
			comparison_filesize = boost::lexical_cast<long long>(pNode->text);

			pNodeMap = pNode->attributes;
			if(pNodeMap == NULL)
				return al::set_ptr();
			
			pNode = pNodeMap->getNamedItem(_bstr_t(XML_ATTRIBUTE_COMPARITOR));
			if(pNode == NULL)
				return al::set_ptr();

			filesize_comparitor = pNode->text;
		}

		// Process optional lastmodified child element
		pNodeList = pElement->getElementsByTagName(_bstr_t(XML_ELEMENT_LASTMODIFIED));
		if(pNodeList == NULL)
			return al::set_ptr();

		if(pNodeList->length > 0)
		{
			lastmodified_option = true;

			pNode = pNodeList->item[0];

			// COMPARITOR
			pNodeMap = pNode->attributes;
			if(pNodeMap == NULL)
				return al::set_ptr();
			
			MSXML2::IXMLDOMNodePtr pNode2 = pNodeMap->getNamedItem(_bstr_t(XML_ATTRIBUTE_COMPARITOR));
			if(pNode2 == NULL)
				return al::set_ptr();

			datetime_comparitor = pNode2->text;

			// YEAR
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_YEAR);
			if(pNode2 == NULL)
				return al::set_ptr();

			int year = boost::lexical_cast<int>(pNode2->text);
			if(year < 1900)
				return al::set_ptr();

			comparison_datetime.tm_year = year - 1900;

			// MONTH
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_MONTH);
			if(pNode2 == NULL)
				return al::set_ptr();

			comparison_datetime.tm_mon = boost::lexical_cast<int>(pNode2->text);

			// DAY
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_DAY);
			if(pNode2 == NULL)
				return al::set_ptr();

			comparison_datetime.tm_mday = boost::lexical_cast<int>(pNode2->text);

			// HOUR
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_HOUR);
			if(pNode2 == NULL)
				return al::set_ptr();

			comparison_datetime.tm_hour = boost::lexical_cast<int>(pNode2->text);

			// MINUTE
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_MINUTE);
			if(pNode2 == NULL)
				return al::set_ptr();

			comparison_datetime.tm_min = boost::lexical_cast<int>(pNode2->text);

			// SECOND
			pNode2 = pNode->selectSingleNode(XML_ATTRIBUTE_SECOND);
			if(pNode2 == NULL)
				return al::set_ptr();

			comparison_datetime.tm_sec = boost::lexical_cast<int>(pNode2->text);
			comparison_datetime.tm_isdst = -1;
		}

		if(!filesize_option && !lastmodified_option)
			return query(root_folder, filename_regex, recurse).run();
		else if(filesize_option && !lastmodified_option)
			return query(root_folder, filename_regex, recurse, filesize_comparitor, comparison_filesize).run();
		else if(!filesize_option && lastmodified_option)
			return query(root_folder, filename_regex, recurse, datetime_comparitor, comparison_datetime).run();
		else if(filesize_option && lastmodified_option)
			return query(root_folder, filename_regex, recurse, filesize_comparitor, comparison_filesize, datetime_comparitor, comparison_datetime).run();
	}
	catch(const boost::bad_lexical_cast& e)
	{
		std::cerr << "Exception: " << e.what() << std::endl;
		return al::set_ptr();
	}
	catch(_com_error& e)
	{
		std::cerr << "COM Exception: " << CT2CA(e.ErrorMessage()) << std::endl;
		return al::set_ptr();
	}

#else /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	xmlDocPtr	pXMLDoc	= NULL;
	xmlNodePtr	pNode	= NULL;
	try
	{
		// Create XML DOM and load up with string passed in
		pXMLDoc = xmlParseMemory(query_xml.c_str(), query_xml.length());
		if(pXMLDoc == NULL)
			return al::set_ptr();

		// Get the root element
		pNode = xmlDocGetRootElement(pXMLDoc);
		if(		(pNode == NULL)
			||	(pNode->type != XML_ELEMENT_NODE) )
			return al::set_ptr();

		// Check root elements name
		if(xmlStrcmp(pNode->name, xmlCharStrdup(XML_ROOT_ELEMENT_NAME)) != 0)
			return al::set_ptr();

		// Read all the attributes
		// searchfolder attribute
		xmlChar* attrib = xmlGetProp(pNode, xmlCharStrdup(XML_ATTRIBUTE_SEARCHFOLDER));
		if(attrib == NULL)
			return al::set_ptr();

		fs::path root_folder = fs::path((const char*)attrib, fs::native);
		xmlFree(attrib);

		// recurse attribute
		attrib = xmlGetProp(pNode, xmlCharStrdup(XML_ATTRIBUTE_RECURSE));
		if(attrib == NULL)
			return al::set_ptr();

		bool recurse = false;
		if(attrib != NULL)
		{
			if(xmlStrcasecmp(xmlCharStrdup(TRUE_STRING), attrib) == 0)
				recurse = true;
			else if(xmlStrcasecmp(xmlCharStrdup(FALSE_STRING), attrib) == 0)
				recurse = false;
			else
				return al::set_ptr();
		}

		xmlFree(attrib);

		// filenameregex attribute
		attrib = xmlGetProp(pNode, xmlCharStrdup(XML_ATTRIBUTE_FILENAMEREGEX));
		if(attrib == NULL)
			return al::set_ptr();

		opl::string filename_regex((const char*)attrib);
		xmlFree(attrib);

		// How handle any optional child elements...
		opl::string filesize_comparitor;
		opl::string datetime_comparitor;
		std::tm comparison_datetime;
		long long comparison_filesize	= 0;
		bool filesize_option			= false;
		bool lastmodified_option		= false;

		for(pNode = pNode->children; 
			pNode != NULL; 
			pNode = pNode->next)
		{
			if(pNode->type == XML_ELEMENT_NODE)
			{
				if(xmlStrcmp(xmlCharStrdup(XML_ELEMENT_FILESIZE), pNode->name) == 0)
				{
					filesize_option = true;

					if(		(pNode->properties != NULL)
						&&	(pNode->properties->type == XML_ATTRIBUTE_NODE)
						&&	(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_COMPARITOR), pNode->properties->name) == 0)
						&&	(pNode->properties->children != NULL)
						&&	(pNode->properties->children->type == XML_TEXT_NODE) )
					{
						filesize_comparitor = opl::string((const char*)pNode->properties->children->content);
					}

					if(		(pNode->children != NULL)
						&&	(pNode->children->type == XML_TEXT_NODE ) )
					{
						comparison_filesize = boost::lexical_cast<long long>((const char*)pNode->children->content);
					}
				}
				else if(xmlStrcmp(xmlCharStrdup(XML_ELEMENT_LASTMODIFIED), pNode->name) == 0)
				{
					lastmodified_option = true;

					if(		(pNode->properties != NULL)
						&&	(pNode->properties->type == XML_ATTRIBUTE_NODE)
						&&	(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_COMPARITOR), pNode->properties->name) == 0)
						&&	(pNode->properties->children != NULL)
						&&	(pNode->properties->children->type == XML_TEXT_NODE) )
					{
						datetime_comparitor = opl::string((const char*)pNode->properties->children->content);
					}

					for(xmlNodePtr pChildNode = pNode->children; 
						pChildNode != NULL;
						pChildNode = pChildNode->next)
					{
						if(		(pChildNode->type == XML_ELEMENT_NODE)
							&&	(pChildNode->children != NULL)
							&&	(pChildNode->children->type == XML_TEXT_NODE) )
						{
							if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_YEAR), pChildNode->name) == 0)
							{
								int year = boost::lexical_cast<int>(pChildNode->children->content);
								if(year < 1900)
									return al::set_ptr();

								comparison_datetime.tm_year = year - 1900;
							}
							else if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_MONTH), pChildNode->name) == 0)
							{
								comparison_datetime.tm_mon = boost::lexical_cast<int>(pChildNode->children->content);
							}
							else if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_DAY), pChildNode->name) == 0)
							{
								comparison_datetime.tm_mday = boost::lexical_cast<int>(pChildNode->children->content);;
							}
							else if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_HOUR), pChildNode->name) == 0)
							{
								comparison_datetime.tm_hour = boost::lexical_cast<int>(pChildNode->children->content);;
							}
							else if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_MINUTE), pChildNode->name) == 0)
							{
								comparison_datetime.tm_min = boost::lexical_cast<int>(pChildNode->children->content);;
							}
							else if(xmlStrcmp(xmlCharStrdup(XML_ATTRIBUTE_SECOND), pChildNode->name) == 0)
							{
								comparison_datetime.tm_sec = boost::lexical_cast<int>(pChildNode->children->content);;
							}
						}
					}
				}
			}
		}

		xmlFreeDoc(pXMLDoc);
		xmlCleanupParser();

		if(!filesize_option && !lastmodified_option)
			return query(root_folder, filename_regex, recurse).run();
		else if(filesize_option && !lastmodified_option)
			return query(root_folder, filename_regex, recurse, filesize_comparitor, comparison_filesize).run();
		else if(!filesize_option && lastmodified_option)
			return query(root_folder, filename_regex, recurse, datetime_comparitor, comparison_datetime).run();
		else if(filesize_option && lastmodified_option)
			return query(root_folder, filename_regex, recurse, filesize_comparitor, comparison_filesize, datetime_comparitor, comparison_datetime).run();
	}
	catch(const boost::bad_lexical_cast& e)
	{
		xmlFreeDoc(pXMLDoc);
		xmlCleanupParser();

		std::cerr << "Exception: " << e.what() << std::endl;
		return al::set_ptr();
	}

#endif ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}

} } } }
