// query.cpp
//
// Copyright (C) 2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef _MSC_VER
#	pragma warning ( push )
#	pragma warning ( disable : 4251 4512 )
#endif

#include <iostream>
#include <boost/filesystem/exception.hpp>
#include <boost/regex.hpp>
#include <openassetlib/al/al.hpp>
#include <openassetlib/al/set.hpp>
#include <openassetlib/plugins/filesystem_storage_plugin/filesystem_storage_plugin.hpp>
#include <openassetlib/plugins/filesystem_storage_plugin/query.hpp>

#ifdef _MSC_VER
#	pragma warning ( pop )
#endif

namespace olib { namespace openassetlib { namespace plugins { namespace filesystem {

query::query(const fs::path& root_folder, const opl::string& filename_regex, bool recurse)
	:	root_folder_(root_folder)
	,	filename_regex_(filename_regex)
	,	recurse_(recurse)
	,	comparison_filesize_(0)
	,	filesize_comparitor_("==")
	,	datetime_comparitor_("==")
{
	comparison_datetime_.tm_year	= 0;
	comparison_datetime_.tm_mon		= 0;
	comparison_datetime_.tm_mday	= 0;
	comparison_datetime_.tm_hour	= 0;
	comparison_datetime_.tm_min		= 0;
	comparison_datetime_.tm_sec		= 0;
	comparison_datetime_.tm_isdst	= -1;
}

query::query(const fs::path& root_folder, const opl::string& filename_regex, bool recurse, const opl::string& filesize_comparitor, long long comparison_filesize)
	:	root_folder_(root_folder)
	,	filename_regex_(filename_regex)
	,	recurse_(recurse)
	,	comparison_filesize_(comparison_filesize)
	,	filesize_comparitor_(filesize_comparitor)
	,	datetime_comparitor_("==")
{
	comparison_datetime_.tm_year	= 0;
	comparison_datetime_.tm_mon		= 0;
	comparison_datetime_.tm_mday	= 0;
	comparison_datetime_.tm_hour	= 0;
	comparison_datetime_.tm_min		= 0;
	comparison_datetime_.tm_sec		= 0;
	comparison_datetime_.tm_isdst	= -1;
}

query::query(const fs::path& root_folder, const opl::string& filename_regex, bool recurse, const opl::string& datetime_comparitor, const std::tm& comparison_datetime)
	:	root_folder_(root_folder)
	,	filename_regex_(filename_regex)
	,	recurse_(recurse)
	,	comparison_filesize_(0)
	,	filesize_comparitor_("==")
	,	datetime_comparitor_(datetime_comparitor)
{
	comparison_datetime_.tm_year	= comparison_datetime.tm_year;
	comparison_datetime_.tm_mon		= comparison_datetime.tm_mon;
	comparison_datetime_.tm_mday	= comparison_datetime.tm_mday;
	comparison_datetime_.tm_hour	= comparison_datetime.tm_hour;
	comparison_datetime_.tm_min		= comparison_datetime.tm_min;
	comparison_datetime_.tm_sec		= comparison_datetime.tm_sec;
	comparison_datetime_.tm_isdst	= -1;
}

query::query(const fs::path& root_folder, const opl::string& filename_regex, bool recurse, const opl::string& filesize_comparitor, long long comparison_filesize, const opl::string& datetime_comparitor, const std::tm& comparison_datetime)
	:	root_folder_(root_folder)
	,	filename_regex_(filename_regex)
	,	recurse_(recurse)
	,	comparison_filesize_(comparison_filesize)
	,	filesize_comparitor_(filesize_comparitor)
	,	datetime_comparitor_(datetime_comparitor)
{
	comparison_datetime_.tm_year	= comparison_datetime.tm_year;
	comparison_datetime_.tm_mon		= comparison_datetime.tm_mon;
	comparison_datetime_.tm_mday	= comparison_datetime.tm_mday;
	comparison_datetime_.tm_hour	= comparison_datetime.tm_hour;
	comparison_datetime_.tm_min		= comparison_datetime.tm_min;
	comparison_datetime_.tm_sec		= comparison_datetime.tm_sec;
	comparison_datetime_.tm_isdst	= -1;
}

query::~query(void)
{
}

al::set_ptr query::run()
{
	try
	{
		if(	    root_folder_.empty()
			|| !root_folder_.is_complete()
			|| !root_folder_.has_root_path()
			|| !fs::exists(root_folder_)
			|| !fs::is_directory(root_folder_))
		{
			return al::set_ptr();
		}
		else
		{
			al::set_ptr set = al::set::create();
			if(!search(root_folder_, set))
				return al::set_ptr();

			return set;
		}
	}
	catch(const fs::filesystem_error& e)
	{
		std::cerr << "Exception: " << e.what() << std::endl;
		return al::set_ptr();
	}
}

bool query::search(const fs::path& directory, al::set_ptr set)
{
	fs::directory_iterator end_itr; // default construction yields past-the-end
	for(fs::directory_iterator itr(directory);
		itr != end_itr;
		++itr )
	{
		if(is_directory(*itr))
		{
			if(recurse_)
			{
				if(!search(*itr, set))
					return false;
			}
		}
		else
		{
			opl::string filename = itr->leaf();

			if(!boost::regex_match(filename.c_str(), boost::regex(filename_regex_.c_str())))
				continue;

			boost::intmax_t	filesize = fs::file_size(*itr);

			if(comparison_filesize_ != 0)
			{
				if(filesize_comparitor_.compare("<") == 0)
				{
					if(filesize >= comparison_filesize_)
						continue;
				}
				else if(filesize_comparitor_.compare("<=") == 0)
				{
					if(filesize > comparison_filesize_)
						continue;
				}
				else if(filesize_comparitor_.compare(">") == 0)
				{
					if(filesize <= comparison_filesize_)
						continue;
				}
				else if(filesize_comparitor_.compare(">=") == 0)
				{
					if(filesize < comparison_filesize_)
						continue;
				}
				else if(filesize_comparitor_.compare("==") == 0)
				{
					if(filesize != comparison_filesize_)
						continue;
				}
				else if(filesize_comparitor_.compare("!=") == 0)
				{
					if(filesize == comparison_filesize_)
						continue;
				}
				else
				{
					std::cerr << "ERROR: unrecognized comparitor" << std::endl;
					return false;
				}
			}

			std::time_t t = fs::last_write_time(*itr);
			if(comparison_datetime_.tm_year != 0)
			{
				time_t comparison_time = mktime(&comparison_datetime_);
				if(comparison_time != (time_t)-1)
				{
					double elapsed_time = difftime(t, comparison_time);

					if(datetime_comparitor_.compare("<") == 0)
					{
						if(elapsed_time >= 0)
							continue;
					}
					else if(datetime_comparitor_.compare("<=") == 0)
					{
						if(elapsed_time > 0)
							continue;
					}
					else if(datetime_comparitor_.compare(">") == 0)
					{
						if(elapsed_time <= 0)
							continue;
					}
					else if(datetime_comparitor_.compare(">=") == 0)
					{
						if(elapsed_time < 0)
							continue;
					}
					else if(datetime_comparitor_.compare("==") == 0)
					{
						if(elapsed_time != 0)
							continue;
					}
					else if(datetime_comparitor_.compare("!=") == 0)
					{
						if(elapsed_time == 0)
							continue;
					}
					else
					{
						std::cerr << "ERROR: unrecognized comparitor" << std::endl;
						return false;
					}
				}
			}

			std::tm* ptm = localtime(&t);

			// create the asset and add it to the set
			al::asset_ptr asset = al::asset::create();

			asset->set("filename", filename);
			asset->set("location", directory.native_file_string());
			asset->set("filesize", filesize);
			asset->set("last_modified", opl::string(::asctime(ptm)));

			set->push_back(asset);
		}
	}

	return true;
}

} } } }
