
// openpluginlib - A plugin interface to openlibraries.

// Copyright (C) 2005 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <algorithm>
#include <cassert>
#include <functional>

#include <boost/filesystem/path.hpp>
#include <boost/regex.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/registry.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>
#include <openpluginlib/pl/shader_manager.hpp>

namespace fs = boost::filesystem;

namespace olib { namespace openpluginlib {

namespace
{
	void reflib( int init, const string& lookup_path = "" )
	{
		static long refs = 0;

		assert( refs >= 0 && L"openpluginlib::refinit: refs is negative." );

		detail::registry& el_reg = detail::registry::instance( );

		if( init > 0 && ++refs )
		{

			if( refs == 1 )
			{
#	ifdef WIN32
				el_reg.insert( plugins_path( ) );
#	else
				el_reg.insert( OPENIMAGELIB_PLUGINS );
				el_reg.insert( OPENMEDIALIB_PLUGINS );
				el_reg.insert( OPENOBJECTLIB_PLUGINS );
				el_reg.insert( OPENASSETLIB_PLUGINS );
#	endif

#	ifdef HAVE_CG_RUNTIME
				shader_manager_instance( ).create_Cg_context( );
#	endif
			}

			if( !lookup_path.empty( ) )
				el_reg.insert( lookup_path );
		}
		else if( init < 0 && --refs == 0 )
		{
			if( !lookup_path.empty( ) )
				el_reg.remove( );
			else
				el_reg.clear( );

#	ifdef HAVE_CG_RUNTIME				
			shader_manager_instance( ).destroy_Cg_context( );
#	endif
		}
	}
	
	boost::recursive_mutex mutex;
}

namespace detail
{
	namespace
	{
		class is_match : public std::unary_function<wstring, bool>
		{
		public:
			explicit is_match( const wstring& to_match )
				: to_match_( to_match )
			{ }

			bool operator( )( const wstring& expr )
			{
				boost::wregex ex( expr.c_str( ), boost::wregex::extended | boost::wregex::icase );
				if( boost::regex_match( to_match_.c_str( ), ex ) )
					return true;

				return false;
			}

		private:
			const wstring& to_match_;
		};

		bool if_matches_expression( const wstring& to_match, const std::vector<wstring>& expression )
		{
			typedef std::vector<wstring>::const_iterator const_iterator;

			const_iterator I = std::find_if( expression.begin( ), expression.end( ), is_match( to_match ) );

			return I != expression.end( );
		}

		void null_delete( void* )
		{ }
	}

	opl_ptr discover_query_impl::plugin_proxy::create_plugin( const string& options ) const
	{
		if( !item_.resolver.dlopened )
			acquire_shared_symbols( item_.resolver, item_.filename );

		if( item_.resolver.dlopened )
		{
			if( item_.resolver.init( ) )
			{
				openplugin* plugin = 0;

				item_.resolver.create_plugin( options.c_str( ), &plugin );

				if( plugin ) return opl_ptr( plugin, item_.resolver.destroy_plugin );
			}
		}

		return opl_ptr( static_cast<openplugin*>( 0 ), null_delete );
	}
	
	bool discover_query_impl::operator( )( const wstring& libname, const wstring& type, const wstring& to_match )
	{
		typedef detail::registry::container						db_container;
		typedef db_container::const_iterator					db_const_iterator;
		typedef db_container::iterator							db_iterator;
		typedef std::pair<db_const_iterator, db_const_iterator>	query_iterator_pair;
		typedef boost::recursive_mutex							mutex;

		mutex mtx;

		const db_container& el_reg_db = detail::registry::instance( ).get_db( );

		query_iterator_pair pair;
		if( libname.empty( ) )
		{
			pair.first  = el_reg_db.begin( );
			pair.second = el_reg_db.end( );
		}
		else
		{
			pair = el_reg_db.equal_range( libname );
		}

		while( pair.first != pair.second )
		{
			const wstring& item_type					= ( *pair.first ).second.type;
			const std::vector<wstring>& item_extension	= ( *pair.first ).second.extension;

			if( ( type.empty( ) || type == item_type ) && ( to_match.empty( ) || if_matches_expression( to_match, item_extension ) ) )
				plugins_.push_back( ( *pair.first ).second );

			++pair.first;
		}
		
		sort<highest_merit_sort>( );

		return true;
	}
}

bool init( const string& lookup_path )
{
	boost::recursive_mutex::scoped_lock lock( mutex );

	reflib( 1, lookup_path );

	return true;
}

bool uninit( )
{
	boost::recursive_mutex::scoped_lock lock( mutex );

	reflib( -1 );

	return true;
}

#ifdef WIN32
namespace
{
	string get_registry_key( const wstring& key_str )
	{
		const int allocsiz = 512;

		TCHAR buf[ allocsiz ];
		DWORD bufsiz = allocsiz;

		string value;
		
		HKEY key;
		if( RegOpenKeyEx( HKEY_LOCAL_MACHINE, L"Software\\OpenLibraries", 0, KEY_READ, &key ) == ERROR_SUCCESS )
		{
			if( RegQueryValueEx( key, key_str.c_str( ), 0, NULL, ( LPBYTE ) buf, &bufsiz ) == ERROR_SUCCESS )
				value = to_string( buf );

			RegCloseKey( key );
		}

		return value;
	}
}

string plugins_path( )
{
#ifndef NDEBUG
	return get_registry_key( L"PluginsDirDebug" );
#else
	return get_registry_key( L"PluginsDirRelease" );
#endif	
}

string kernels_path( )
{ return get_registry_key( L"KernelsDir" ); }

string shaders_path( )
{ return get_registry_key( L"ShadersDir" ); }
#endif 

} }
